<?

/*
	Show sudoku's from database as PDF

        (C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License 
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

include('utilities.php');

function show_sudoku($xmin, $xmax, $ymin, $ymax, $nx, $ny, $serial, $problem, $solution, $showsolution, $showpossibilities, $txt, $solpage)
{
  global $SYMBOLLIST;
  $p = "";
  $nxny = $nx*$ny;

  if(($xmax-$xmin)>($ymax-$ymin))
  {
    $size = $ymax-$ymin;
  }
  else
  {
    $size = $xmax-$xmin;
  }
  $size = $size*0.8;
  $p .= "gsave ".(($xmax-$xmin)/2-$size/2+$xmin)." ".(($ymax-$ymin)/2-$size/2+$ymin)." translate\n";
  $cellsize = $size/$nxny;
  $fontsize = $cellsize*0.9;
  $hintfontsize = $fontsize/max($nx,$ny);
  $headlinefontsize = $size/25;
  $subtextfontsize = $size/35;
  $thinwidth = $cellsize/40;
  $thickwidth = $thinwidth*3;
  $p .= "/Helvetica-ISOLatin1 findfont ".$headlinefontsize." scalefont setfont\n";
  $p .= ($size/2)." ".($size+3*$headlinefontsize)." moveto (Sudoku nr. ".$serial.") show4\n";

  $p .= "/Helvetica-ISOLatin1 findfont ".$fontsize." scalefont setfont\n";
  $y = $size;
  $i=0;
  if($showsolution)
    $probsol = $solution;
  else
    $probsol = $problem;

  for($row=0; $row<$nxny; $row++)
  {
    if($row>0)
    {
      $p .= "0 0 0 setrgbcolor\n";
      $width = ($row%$ny==0)?$thickwidth:$thinwidth;
      $p .= ($width)." setlinewidth\n";
      $p .= "newpath\n";
      $p .= $y." 0 moveto\n";
      $p .= $y." ".$size." lineto\n";
      $p .= "stroke\n";
      $width = ($row%$nx==0)?$thickwidth:$thinwidth;
      $p .= ($width)." setlinewidth\n";
      $p .= "newpath\n";
      $p .= "0 ".$y." moveto\n";
      $p .= $size." ".$y." lineto\n";
      $p .= "stroke\n";
    }
    $x=0;
    for($col=0; $col<$nxny; $col++)
    {
      $c = substr($probsol, $i, 1);
      if($c!="0")
      {
	$p .= (substr($problem, $i, 1)=="0"?"1":"0")." 0 0 setrgbcolor\n";
	$p .= ($x+$cellsize/2)." ".($y-$cellsize/2)." moveto\n";
	$p .= "(".$c.") show5\n";
      }
      else if($showpossibilities)
      {
	$p .= "/Helvetica-ISOLatin1 findfont ".$hintfontsize." scalefont setfont\n";
	find_possibilities($nx, $ny, $row, $col, $problem, $possibilities);
	for($i1=0;$i1<$nx;$i1++)
	for($j1=0;$j1<$ny;$j1++)
	{
	  $value = $i1*$ny+$j1+1;
	  if($possibilities[$value])
	  {
	    $p .= ($x+$cellsize/$ny*($j1+0.5))." ".($y-$cellsize/$nx*($i1+0.5))." moveto\n";
	    $p .= "(".substr($SYMBOLLIST,$value,1).") show5\n";
	  }
	}
	$p .= "/Helvetica-ISOLatin1 findfont ".$fontsize." scalefont setfont\n";
      }
      $i++;
      $x += $cellsize;
    }
    $y -= $cellsize;
  }
  $p .= $thickwidth." setlinewidth\n";
  $p .= "0 0 0 setrgbcolor\n";
  $p .= "newpath\n";
  $p .= "0 0 moveto\n";
  $p .= "0 ".$size." lineto\n";
  $p .= $size." ".$size." lineto\n";
  $p .= $size." 0 lineto\n";
  $p .= "closepath\n";
  $p .= "stroke\n";
  if(!$showsolution)
  {
    $p .= "/Helvetica-ISOLatin1 findfont ".$subtextfontsize." scalefont setfont\n";
    $y = -3*$subtextfontsize;
    if($txt != "")
    {
      $p .= ($size/2)." ".$y." moveto (".$txt.") show4\n";
      $y -= 1.5*$subtextfontsize;
    }
    $p .= ($size/2)." ".$y." moveto (Lsning side ".$solpage.") show4\n";
    $y -= 1.5*$subtextfontsize;
  }


  $p .= "grestore\n";
  return $p;
}

sudoku_db();

$maxnumber=100;
$solnx=2;
$solny=3;

if(isset($action))
{
  /*
  	Generate the following arrays:

	$serials	Serial number of all problems
	$pages		Indexed as $serials: Page number on which the problem is printed
	$problems
	$solutions
	$nxs
	$nys
	$texts
	$solpages	Page no of solution
	$cattext	Category text
	$catfirst	Index into $serials of first

  */
  $ntotal=0;
  $pageno=2;
  if(isset($number))
  {
    while(list($nx,$nylist)=each($number))
    {
      while(list($ny,$levels)=each($nylist))
      {
        while(list($level,$n)=each($levels))
	{
	  if($n>0)
	  {
	    if(($n+$ntotal)>$maxnumber) $n=$maxnumber-$ntotal;
	    if($n>0)
	    {
	      $res1 = mysql_query($q="select * from sudoku where ".
	              "nx=".$nx." and ".
		      "ny=".$ny." and ".
		      "level=".$level." and ".
		      "(1=0".
		      (isset($symmetrical[0])?" or symmetrical=0":"").
		      (isset($symmetrical[1])?" or symmetrical=1":"").
		      (isset($symmetrical[2])?" or symmetrical=2":"").
		      ") order by lastused".(isset($fewfilled)?",ndigits":"")." limit ".$n);
	      $n = mysql_num_rows($res1);
	      if($n>0)
	      {
	        $cattext[] = $nx."x".$ny." ".level_text($level);
		$catfirst[] = count($serials);
		while($row1=mysql_fetch_assoc($res1))
		{
		  $serials[] = $row1["serial"];
		  $nxs[] = $row1["nx"];
		  $nys[] = $row1["ny"];
		  $texts[] = "Kategori: ".level_text($level)." Symmetri: ".symmetrical_text($row1["symmetrical"]);
		  $problems[] = $row1["problem"];
		  $solutions[] = $row1["solution"];
		  $pages[] = $pageno;
		  $pageno++;
		  $res2 = mysql_query("update sudoku set lastused=now() where serial=".$row1["serial"]);
		}
	      }
	    }
            $ntotal += $n;
	  }
	}
      }
    }
  }
  if($ntotal==0)
  {
    echo "<HTML>\n";
    echo "<HEAD>\n";
    echo "  <META HTTP-EQUIV=\"content-type\" CONTENT=\"TEXT/HTML; CHARSET=ISO-8859-1\">\n";
    echo "  <TITLE>Sudoku opgavest</TITLE>\n";
    echo "  <link rel=\"stylesheet\" type=\"text/css\" href=\"sudoku.css\" title=\"Default\">\n";
    echo "</HEAD>\n";
    echo "<body>\n";

    echo "<h1>Fejl!</h1>\n";

    echo "<p>Der er ingen opgaver at vise. Enten har du ikke fyldt tal i felterne, eller du har bedt om opgavetyper, som ikke findes i databasen. Klik <A HREF=\"javascript:history.back(-1)\">tilbage</a> og udfyld felterne igen.</p>\n";

    echo "</body>\n";
    echo "</html>\n";
    exit;
  }
  /* Walk through all problems and calculate page number of solution */
  reset($serials);
  $i=0;
  while($serial=each($serials))
  {
    $solpages[] = $pageno;
    $i++;
    if($i>=($solnx*$solny))
    {
      $i=0;
      $pageno++;
    }
  }
  if($i==0)
  {
    $pageno--;
  }
  $xmax = 595;
  $ymax = 842;
  $p = "%!PS-Adobe-2.1\n";
  $p .= "%%Creator: http://www.lemo.dk/sudoku\n";
  $p .= "%%Pages: ".$pageno."\n";
  $p .= "%%DocumentMedia: A4 ".$xmax." ".$ymax." 0 white ()\n";
  $p .= "%%BoundingBox: 0 0 ".$xmax." ".$ymax."\n";
  $p .= "%%Orientation: Portrait\n";
  $p .= "%%EndComments\n";
  $p .= "[{\n";
  $p .= "%%BeginFeature: *PageSize A4\n";
  $p .= "<</PageSize[".$xmax." ".$ymax."]/ImagingBBox null>>setpagedevice\n";
  $p .= "%%EndFeature\n";
  $p .= "} stopped cleartomark\n";
  $p .= "mark\n";
  $p .= "/ISOLatin1Encoding \n";
  $p .= "8#000 1 8#054 {StandardEncoding exch get} for \n";
  $p .= "/minus \n";
  $p .= "8#056 1 8#217 {StandardEncoding exch get} for \n";
  $p .= "/dotlessi \n";
  $p .= "8#301 1 8#317 {StandardEncoding exch get} for \n";
  $p .= "/space /exclamdown /cent /sterling /currency /yen /brokenbar /section \n";
  $p .= "/dieresis /copyright /ordfeminine /guillemotleft /logicalnot /hyphen \n";
  $p .= "/registered /macron /degree /plusminus /twosuperior /threesuperior /acute \n";
  $p .= "/mu /paragraph /periodcentered /cedilla /onesuperior /ordmasculine \n";
  $p .= "/guillemotright /onequarter /onehalf /threequarters /questiondown /Agrave \n";
  $p .= "/Aacute /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla /Egrave /Eacute\n";
  $p .= "/Ecircumflex /Edieresis /Igrave /Iacute /Icircumflex /Idieresis /Eth /Ntilde\n";
  $p .= "/Ograve /Oacute /Ocircumflex /Otilde /Odieresis /multiply /Oslash /Ugrave \n";
  $p .= "/Uacute /Ucircumflex /Udieresis /Yacute /Thorn /germandbls /agrave /aacute \n";
  $p .= "/acircumflex /atilde /adieresis /aring /ae /ccedilla /egrave /eacute \n";
  $p .= "/ecircumflex /edieresis /igrave /iacute /icircumflex /idieresis /eth /ntilde\n";
  $p .= "/ograve /oacute /ocircumflex /otilde /odieresis /divide /oslash /ugrave \n";
  $p .= "/uacute /ucircumflex /udieresis /yacute /thorn /ydieresis \n";
  $p .= "/ISOLatin1Encoding where not {256 array astore def} if \n";
  $p .= " cleartomark \n";
  $p .= "/makeISOEncoded \n";
  $p .= "{ findfont /curfont exch def \n";
  $p .= "/newfont curfont maxlength dict def  \n";
  $p .= "/ISOLatin1 (-ISOLatin1) def\n";
  $p .= "/curfontname curfont /FontName get dup length string cvs def \n";
  $p .= "/newfontname curfontname length ISOLatin1 length add string \n";
  $p .= "dup 0 curfontname putinterval \n";
  $p .= "dup curfontname length ISOLatin1 putinterval \n";
  $p .= "def \n";
  $p .= "curfont   \n";
  $p .= "{ exch dup /FID ne  \n";
  $p .= "{ dup /Encoding eq  \n";
  $p .= "{ exch pop ISOLatin1Encoding exch }  \n";
  $p .= "if dup /FontName eq  \n";
  $p .= "{ exch pop newfontname exch }  \n";
  $p .= "if exch newfont 3 1 roll put  \n";
  $p .= "}  \n";
  $p .= "{ pop pop }  \n";
  $p .= "ifelse  \n";
  $p .= "  }  \n";
  $p .= "  forall newfontname newfont definefont \n";
  $p .= "} def \n";
  $p .= "/Times-Roman makeISOEncoded pop \n";
  $p .= "/Times-Bold makeISOEncoded pop \n";
  $p .= "/Times-Italic makeISOEncoded pop \n";
  $p .= "/Times-BoldItalic makeISOEncoded pop \n";
  $p .= "/Helvetica makeISOEncoded pop \n";
  $p .= "/Helvetica-Bold makeISOEncoded pop \n";
  $p .= "/Helvetica-Oblique makeISOEncoded pop \n";
  $p .= "/Helvetica-BoldOblique makeISOEncoded pop \n";
  $p .= "/Courier makeISOEncoded pop \n";
  $p .= "/Courier-Bold makeISOEncoded pop \n";
  $p .= "/Courier-Oblique makeISOEncoded pop \n";
  $p .= "/Courier-BoldOblique makeISOEncoded pop \n";

  $p .=  "/sbound {dup gsave 0 0 moveto true charpath \n";
  $p .=  "         flattenpath pathbbox grestore \n";
  $p .=  "       /symax exch def /sxmax exch def \n";
  $p .=  "       /symin exch def /sxmin exch def} def \n";
  /*
      anchor points:


      3--------6--------9
      |                 |
      2        5        8
      |                 |
      1--------4--------7

  */
  $p .=  "/show1 {sbound sxmin neg symin neg rmoveto show} def\n";
  $p .=  "/show2 {sbound sxmin neg symin symax sub 2 div rmoveto show} def\n";
  $p .=  "/show3 {sbound sxmin neg symax neg rmoveto show} def\n";
  $p .=  "/show4 {sbound sxmin sxmax sub 2 div sxmin sub symin neg rmoveto show} def\n";
  $p .=  "/show5 {sbound sxmin sxmax sub 2 div sxmin sub symin symax sub 2 div rmoveto show} def\n";
  $p .=  "/show6 {sbound sxmin sxmax sub 2 div sxmin sub symax neg rmoveto show} def\n";
  $p .=  "/show7 {sbound sxmax neg symin neg rmoveto show} def\n";
  $p .=  "/show8 {sbound sxmax neg symin symax sub 2 div rmoveto show} def\n";
  $p .=  "/show9 {sbound sxmax neg symax neg rmoveto show} def\n";

  /* Create front page */

  $p .= "0 0 0 setrgbcolor\n";
  $y = $ymax*0.8;
  $p .= "/Helvetica-ISOLatin1 findfont 50 scalefont setfont\n";
  $p .= ($xmax/2)." ".$y." moveto (".$ntotal." Sudoku opgave".($ntotal==1?"":"r").") show5\n";
  $y -= 72;
  $p .= "/Helvetica-ISOLatin1 findfont 10 scalefont setfont\n";
  $p .= ($xmax/2)." ".$y." moveto (http://www.lemo.dk/sudoku) show5\n";
  $y -= 72;
  $p .= "/Helvetica-ISOLatin1 findfont 10 scalefont setfont\n";
  setlocale(LC_TIME, "da_DK");
  $p .= ($xmax/2)." ".$y." moveto (Fremstillet: ".strftime("%e. %B, %Y, %H:%M:%S").") show5\n";

  $y = $ymax*0.5;
  $p .= "/Helvetica-ISOLatin1 findfont 20 scalefont setfont\n";
  $p .= ($xmax/2)." ".$y." moveto (Indholdsfortegnelse:) show5\n";
  $y -= 60;
  $p .= "/Helvetica-ISOLatin1 findfont 10 scalefont setfont\n";
  for($i=0; $i<count($cattext); $i++)
  {
    $p .= "108 $y moveto (".$cattext[$i].") show2\n";
    $pageno = $pages[$catfirst[$i]];
    $p .= "595 108 sub $y moveto (".$pageno.") show8\n";
    $y -= 15;
  }
  $y -= 15;
  $p .= "108 $y moveto (Lsning".($ntotal==1?"":"er").":) show2\n";
  $pageno = $solpages[0];
  $p .= "595 108 sub $y moveto (".$pageno.") show8\n";
  $p .= "showpage\n";
  $pageno = 2;
  for($index=0; $index<count($serials); $index++)
  {
    $p .= "/Helvetica-ISOLatin1 findfont 10 scalefont setfont\n";
    $p .= ($xmax/2)." 36 moveto (".$pageno.") show4\n";
    $p .= show_sudoku(0,$xmax,0,$ymax,
                      $nxs[$index],$nys[$index],$serials[$index],
		      $problems[$index],$solutions[$index],
		      0,isset($showpossibilities),$texts[$index],$solpages[$index]);
    $p .= "showpage\n";
    $pageno++;
  }
  $i=0;
  $bottommargin = 56;
  for($index=0; $index<count($serials); $index++)
  {
    if($i==0)
    {
      $p .= "/Helvetica-ISOLatin1 findfont 10 scalefont setfont\n";
      $p .= ($xmax/2)." 36 moveto (".$pageno.") show4\n";
    }
    $xsize = $xmax/$solnx;
    $ysize = ($ymax-$bottommargin*2)/$solny;
    $xmin2 = ($i%$solnx)*$xsize;
    $xmax2 = $xmin2+$xsize;
    $ymax2 = $ymax-$ysize*(floor($i/$solnx))-$bottommargin;
    $ymin2 = $ymax2-$ysize;
    $p .= show_sudoku($xmin2,$xmax2,$ymin2,$ymax2,
                      $nxs[$index],$nys[$index],$serials[$index],
		      $problems[$index],$solutions[$index],
		      1,0,$texts[$index],$solpages[$index]);
    $i++;
    if($i == ($solnx*$solny))
    {
      $p .= "showpage\n";
      $pageno++;
      $i=0;
    }
  }
  if($i>0)
  {
    $p .= "showpage\n";
  }

  header("Content-type: application/pdf");
  $tmpfile = tempnam("/tmp","pdf");
  $fd = fopen($tmpfile, "w");
  fwrite($fd, $p);
  fclose($fd);
  passthru("/usr/bin/gs -q -dNOPAUSE -dBATCH -sPAPERSIZE=a4 -sDEVICE=pdfwrite -sOutputFile=- -c save pop -f ".$tmpfile);
  unlink($tmpfile);
}
else
{
  echo "<HTML>\n";
  echo "<HEAD>\n";
  echo "  <META HTTP-EQUIV=\"content-type\" CONTENT=\"TEXT/HTML; CHARSET=ISO-8859-1\">\n";
  echo "  <TITLE>Sudoku opgavest</TITLE>\n";
  echo "  <link rel=\"stylesheet\" type=\"text/css\" href=\"sudoku.css\" title=\"Default\">\n";
  echo "</HEAD>\n";
  echo "<body>\n";

  echo "<h1>Sudoku opgavest</h1>\n";

  echo "<p>Udfyld skemaet forneden med antallet af opgaver i de forskellige strrelser\n";
  echo "og svrhedsgrader, klik p <b>Generr</b>, og du fr en PDF fil med opgaverne\n";
  echo "og lsningerne, lige til at udskrive og tage med p ferie, i toget, etc.</p>\n";

  echo "<p>Der kan maksimalt vises ".$maxnumber." opgaver i hver PDF fil.</p>\n";

  echo "<p><b>Bemrk:</b> Hvis databasen ikke indeholder s mange opgaver af en bestemt type, ";
  echo "bliver antallet reduceret. Hvis der vlges for mange kategorier, kan indholdsfortegnelsen ikke vises i sin helhed.</p>\n";

  echo "<p><hr></p>\n";

  echo "<form method=post action=\"$PHP_SELF\">\n";
  echo "<table border=1>\n";
  echo "<tr><th colspan=6>nskede symmetrier i opgaverne";
  echo "<tr><th colspan=5>".symmetrical_text(0)."<td align=center><input type=checkbox name=symmetrical[0]>\n";
  echo "<tr><th colspan=5>".symmetrical_text(1)."<td align=center><input type=checkbox name=symmetrical[1] checked>\n";
  echo "<tr><th colspan=5>".symmetrical_text(2)."<td align=center><input type=checkbox name=symmetrical[2]>\n";
  echo "<tr><th rowspan=2>Strrelse<th colspan=5>Antal opgaver:<br>Svrhedsgrad\n";
  echo "<tr><th>Meget nem<th>Nem<th>Medium<th>Svr<th>Meget svr\n";

  $res1 = mysql_query("select nx,ny,level from sudoku group by nx,ny,level");

  $lastnx = 999;
  $lastny = 999;
  $lastlevel = 4;
  while($row1 = mysql_fetch_assoc($res1))
  { 
    if($lastnx != $row1["nx"] ||
       $lastny != $row1["ny"])
    {
      for($i=0;$i<(4-$lastlevel);$i++)
      {
	echo "<td>&nbsp;";
      }
      echo "<tr>";
      echo "<td align=center>".$row1["nx"]."x".$row1["ny"];
      $lastlevel = -1;
    }
    for($i=0;$i<($row1["level"]-$lastlevel-1);$i++)
    {
      echo "<td>&nbsp;";
    }
    echo "<td align=center><input name=number[".$row1["nx"]."][".$row1["ny"]."][".$row1["level"]."] size=3>\n";
    $lastnx = $row1["nx"];
    $lastny = $row1["ny"];
    $lastlevel = $row1["level"];
  }
  for($i=0;$i<(4-$lastlevel);$i++)
  {
    echo "<td>&nbsp;";
  }
  echo "<tr><th colspan=5>S f fyldte celler som muligt<td align=center><input type=checkbox name=fewfilled>\n";
  echo "<tr><th colspan=5>Vis mulige kandidater i hver tomme celle<td align=center><input type=checkbox name=showpossibilities>\n";
  echo "<tr><td colspan=3 align=center><input type=reset value=\"Ryd felter\">\n";
  echo "<td colspan=3 align=center><input type=submit name=action value=\"Generr\">\n";
  echo "</table>\n";
  echo "</form>\n";

  echo "</body>\n";
  echo "</html>\n";
}
?>
